from . import sampler
from .topology import Edge, Face, Halfedge, Loop, Shell, Solid as TopoSolid, Topology
from .curve import Circle, Ellipse, Line, BSplineCurve, Other
from .surface import *
from .winding_number import find_surface_uv_for_curve
import numpy as np


class Shape:
    """Represents a geometric part (one part from the CAD model) with geometry and topology assembled."""
    def __init__(self, geometry_data, topology_data, version, spacing=0.02):

        self._geometry_data = self._geometry_data(geometry_data, version)
        self._topology_data = self._topology_data(topology_data, version)
        spacing *= np.linalg.norm(self._geometry_data.bbox[0][1] - self._geometry_data.bbox[0][0])
        self.bbox = self._geometry_data.bbox[0]
        self.vertices = self._geometry_data.vertices

        self._create_2d_trimming_curves(self._geometry_data.curves2d, self._geometry_data.curves3d, spacing)
        self.edges, self.faces, self.halfedges, self.loops, self.shells, self.solids = \
            self._assemble_topology(self._topology_data, self._geometry_data, self.trimming_curves_2d)

    def _create_2d_trimming_curves(self, curves_2d, curves_3d, spacing):
        """
        Create 2D trimming curves.
        """
        self.trimming_curves_2d = []
        for shell in self._topology_data.shells:
            self._process_2d_trimming_curves_for_shell(shell, curves_2d, curves_3d, spacing)


    def _process_2d_trimming_curves_for_shell(self, shell_index, curves2d, curves3d, spacing):

        if isinstance(shell_index, Shell):
            shell = shell_index
        else:
            shell = self._topology_data.shells[shell_index]

        for (face_index, _) in shell.faces:
            face = self._topology_data.faces[face_index]
            self.trimming_curves_2d += (face_index-len(self.trimming_curves_2d)+1)* [None]
            self.trimming_curves_2d[face_index] = []

            surface_index = face.surface
            surface = self._geometry_data.surfaces[surface_index]

            for loop_id in face.loops:
                loop = self._topology_data.loops[loop_id]
                for halfedge_index in loop.halfedges:
                    halfedge = self._topology_data.halfedges[halfedge_index]
                    modified_orientation = halfedge.orientation_wrt_edge
                    if not face.surface_orientation:
                        modified_orientation = not modified_orientation

                    curve3d_index = halfedge.edge
                    curve3d = curves3d[curve3d_index]

                    if hasattr(halfedge, 'curve2d'):
                        curve2d_index = halfedge.curve2d
                        current_curve = curves2d[curve2d_index]
                    elif curve3d is not None:
                        current_curve = curve3d
                    else:
                        continue

                    if current_curve.shape_name == 'Line':
                        n_samples = 2
                    else:
                        length = current_curve.get_length()
                        n_samples = int(length / spacing)

                    if hasattr(halfedge, 'curve2d'):
                        curve2d_index = halfedge.curve2d
                        curve2d = curves2d[curve2d_index]
                        _, closest_surface_uv_values_of_curve = sampler.uniform_sample(curve2d, n_samples, 4, 100)
                        if not modified_orientation:
                            closest_surface_uv_values_of_curve = closest_surface_uv_values_of_curve[::-1]
                    else:
                        surface_uv_values, surface_points = sampler.uniform_sample(surface, n_samples*n_samples, 5, 100)

                        # Sample the curve points to get UV values
                        _, curve_points = sampler.uniform_sample(curve3d, n_samples)

                        if not modified_orientation:
                            curve_points = curve_points[::-1]
                        # Calculate the nearest UV values on the surface for the curve points
                        closest_surface_uv_values_of_curve = find_surface_uv_for_curve(surface_points, surface_uv_values, curve_points)

                    self.trimming_curves_2d[face_index].append(closest_surface_uv_values_of_curve)

    def _assemble_topology(self, topo, geo, trimming_curves):
        edges, faces, halfedges, loops, shells, solids = [], [], [], [], [], []

        # Edges
        for edge in topo.edges:
            edge.curve3d = geo.curves3d[int(edge.curve3d)]
            edges.append(edge)

        # Halfedges
        for halfedge in topo.halfedges:
            c2d = int(halfedge.curve2d)
            halfedge.curve2d = geo.curves2d[c2d] if (c2d is not None and c2d < len(geo.curves2d)) else None

            halfedge.edge = edges[int(halfedge.edge)]

            if halfedge.mates:
                halfedge.mates = topo.halfedges[int(halfedge.mates[0])]
            halfedges.append(halfedge)

        # Loops
        for loop in topo.loops:
            loop.halfedges = [topo.halfedges[int(hid)] for hid in loop.halfedges]
            loops.append(loop)

        # Faces
        for idx, face in enumerate(topo.faces):
            face.surface = geo.surfaces[int(face.surface)]
            face.loops = [topo.loops[int(lid)] for lid in face.loops]

            # safer indexing (in case trimming_curves is shorter)
            face.trimming_curves_2d = trimming_curves[idx] if idx < len(trimming_curves) else None

            faces.append(face)

        # Shells
        for shell in topo.shells:
            shell.faces = [(topo.faces[int(f)], bool(o)) for (f, o) in shell.faces]
            shells.append(shell)

        # Solids
        for solid in topo.solids:
            solid.shells = [topo.shells[int(sid)] for sid in solid.shells]
            solids.append(solid)

        # ---- reverse mappings (same as before but using locals) ----

        # edges -> halfedges
        edgeMap = {}
        for he in halfedges:
            e = he.edge
            edgeMap.setdefault(e, {'halfedges': []})['halfedges'].append(he)
        for e in edgeMap:
            e._halfedges = edgeMap[e]['halfedges']

        # halfedges -> loops
        halfEdgeMap = {}
        for lp in loops:
            for he in lp.halfedges:
                halfEdgeMap.setdefault(he, {'loops': []})['loops'].append(lp)
        for he in halfEdgeMap:
            he.loops = halfEdgeMap[he]['loops']

        # loops -> faces
        loopMap = {}
        for fc in faces:
            for lp in fc.loops:
                loopMap.setdefault(lp, {'faces': []})['faces'].append(fc)
        for lp in loopMap:
            lp.faces = loopMap[lp]['faces']

        # faces -> shells
        faceMap = {}
        for sh in shells:
            for fc, _ in sh.faces:
                faceMap.setdefault(fc, {'shells': []})['shells'].append(sh)
        for fc in faceMap:
            fc.shells = faceMap[fc]['shells']

        # shells -> solids
        if len(solids) > 0:
            shellMap = {}
            for sd in solids:
                for sh in sd.shells:
                    shellMap.setdefault(sh, {'solids': []})['solids'].append(sd)
            for sh in shellMap:
                sh.solids = shellMap[sh]['solids']

        # solids -> parts
        solidMap = {}
        for sd in solids:
            solidMap.setdefault(sd, {'parts': []})['parts'].append(self)

        for sd in solidMap:
            sd.part = solidMap[sd]['parts']


        return edges, faces, halfedges, loops, shells, solids

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented

        def summarize_edges(edges):
            return {
                e.id: (
                    int(e.start_vertex),
                    int(e.end_vertex),
                    e.curve3d,
                )
                for e in edges
            }

        def summarize_halfedges(halfedges):
            summary = {}
            for h in halfedges:
                mate = getattr(h, "mates", None)
                mate_id = mate.id if mate is not None else None

                summary[h.id] = (
                    bool(h.orientation_wrt_edge),
                    h.edge.id,
                    mate_id,
                    h.curve2d,
                )
            return summary

        def summarize_loops(loops):

            return {
                loop.id: tuple(h.id for h in loop.halfedges)
                for loop in loops
            }

        def summarize_faces(faces):
            summary = {}
            for f in faces:
                exact_domain = tuple(np.asarray(f.exact_domain).ravel().tolist())
                loop_ids = tuple(loop.id for loop in f.loops)

                sing_raw = getattr(f, "singularities", None)
                sing_summary = None
                if isinstance(sing_raw, dict):
                    sing_summary = {}
                    for key, inner in sing_raw.items():
                        norm_inner = {}
                        for k2, v in inner.items():
                            if isinstance(v, (np.ndarray, list, tuple)):
                                norm_inner[k2] = np.asarray(v).ravel().tolist()
                            else:
                                norm_inner[k2] = v
                            sing_summary[key] = norm_inner

                summary[f.id] = (
                    exact_domain,
                    bool(getattr(f, "has_singularities", False)),
                    int(getattr(f, "nr_singularities", 0)),
                    int(f.outer_loop),
                    bool(f.surface_orientation),
                    loop_ids,
                    # sing_summary,
                    f.surface,
                )
            return summary

        def summarize_shells(shells):
            summary = {}
            for s in shells:
                faces = tuple((face.id, bool(orient)) for (face, orient) in s.faces)
                summary[s.id] = (
                    bool(s.orientation_wrt_solid),
                    faces,
                )
            return summary

        def summarize_solids(solids):
            summary = {}
            for s in solids:
                shell_ids = tuple(shell.id for shell in s.shells)
                summary[s.id] = shell_ids
            return summary

        return (
            summarize_edges(self.edges) == summarize_edges(other.edges)
            and summarize_halfedges(self.halfedges) == summarize_halfedges(other.halfedges)
            and summarize_loops(self.loops) == summarize_loops(other.loops)
            and summarize_faces(self.faces) == summarize_faces(other.faces)
            and summarize_shells(self.shells) == summarize_shells(other.shells)
            and summarize_solids(self.solids) == summarize_solids(other.solids)
        )

    class _geometry_data:
        def __init__(self, geometry_data, version):
            self.curves2d, self.curves3d, self.surfaces, self.bbox, self.vertices = [], [], [], [], []
            self.__init_geometry(geometry_data, version)

        def _create_3dcurve(self, tmp):
            ctype = int(tmp[0])
            interval = tmp[1:3]
            transform = tmp[len(tmp) - 12:]  # last 16 values are the transformation matrix
            transform = np.array(transform).reshape((3, 4))

            if ctype == 0:  # Line
                res = Line(None,
                                        interval=interval,
                                        location=tmp[3:6],
                                        direction=tmp[6:9],
                                        transform=transform)
            elif ctype == 1:  # Circle
                res = Circle(None,
                                          interval=interval,
                                          location=tmp[3:6],
                                          x_axis=tmp[6:9],
                                          y_axis=tmp[9:12],
                                          z_axis=tmp[12:15],
                                          radius=tmp[15],
                                          transform=transform)
            elif ctype == 2:  # Ellipse
                res = Ellipse(None,
                                           interval=interval,
                                           focus1=tmp[3:6],
                                           focus2=tmp[6:9],
                                           x_axis=tmp[9:12],
                                           y_axis=tmp[12:15],
                                           z_axis=tmp[15:18],
                                           maj_radius=tmp[18],
                                           min_radius=tmp[19],
                                           transform=transform)
            elif ctype == 3:  # BSplineCurve
                degree = int(tmp[3])
                continuity = int(tmp[4])
                rational = bool(tmp[5])
                periodic = bool(tmp[6])
                closed = bool(tmp[7])
                len_poles = int(tmp[8])
                pshape = (int(tmp[9]), int(tmp[10]))
                poles = np.array(tmp[11:11 + len_poles]).reshape(pshape)
                idx = 11 + len_poles
                len_knots = int(tmp[idx])
                knots = tmp[idx + 1:idx + 1 + len_knots]
                idx = idx + 1 + len_knots
                len_weights = int(tmp[idx])
                weights = tmp[idx + 1:idx + 1 + len_weights]

                res = BSplineCurve(None,
                                                interval=interval,
                                                degree=degree,
                                                continuity=continuity,
                                                rational=rational,
                                                periodic=periodic,
                                                closed=closed,
                                                poles=poles,
                                                knots=knots,
                                                weights=weights,
                                                transform=transform)
            elif ctype == 4:  # Other
                res = Other(None, interval=interval, transform=transform)
            else:
                raise ValueError(f"Unknown curve type: {ctype} for curve {tmp}")

            return res

        def _create_surface(self, tmp):
            transform = np.array(tmp[-12:]).reshape((3, 4))
            payload = tmp[:-12]
            stype = int(payload[0])

            trim_domain = np.array(payload[1:5]).reshape((2, 2))
            idx = 5

            if stype == 0:  # Plane
                res = Plane(None,
                             trim_domain=trim_domain,
                             transform=transform,
                             location=np.array(payload[idx:idx + 3]),
                             coefficients=np.array(payload[idx + 3:idx + 7]),
                             x_axis=np.array(payload[idx + 7:idx + 10]),
                             y_axis=np.array(payload[idx + 10:idx + 13]),
                             z_axis=np.array(payload[idx + 13:idx + 16]))

            elif stype == 1:  # Cylinder
                res = Cylinder(None,
                                trim_domain=trim_domain,
                                transform=transform,
                                location=np.array(payload[idx:idx + 3]),
                                radius=payload[idx + 3],
                                coefficients=np.array(payload[idx + 4:-9]),
                                x_axis=np.array(payload[-9:-6]),
                                y_axis=np.array(payload[-6:-3]),
                                z_axis=np.array(payload[-3:]))

            elif stype == 2:  # Cone
                res = Cone(None,
                            trim_domain=trim_domain,
                            transform=transform,
                            location=np.array(payload[idx:idx + 3]),
                            radius=payload[idx + 3],
                            coefficients=np.array(payload[idx + 4:-13]),
                            apex=np.array(payload[-13:-10]),
                            angle=payload[-10],
                            x_axis=np.array(payload[-9:-6]),
                            y_axis=np.array(payload[-6:-3]),
                            z_axis=np.array(payload[-3:]))

            elif stype == 3:  # Sphere
                res = Sphere(None,
                              trim_domain=trim_domain,
                              transform=transform,
                              location=np.array(payload[idx:idx + 3]),
                              radius=payload[idx + 3],
                              coefficients=np.array(payload[idx + 4:-9]),
                              x_axis=np.array(payload[-9:-6]),
                              y_axis=np.array(payload[-6:-3]),
                              z_axis=np.array(payload[-3:]))

            elif stype == 4:  # Torus
                res = Torus(None,
                             trim_domain=trim_domain,
                             transform=transform,
                             location=np.array(payload[idx:idx + 3]),
                             max_radius=payload[idx + 3],
                             min_radius=payload[idx + 4],
                             x_axis=np.array(payload[idx + 5:idx + 8]),
                             y_axis=np.array(payload[idx + 8:idx + 11]),
                             z_axis=np.array(payload[idx + 11:idx + 14]))

            elif stype == 5:  # BSplineSurface
                u_degree, v_degree, continuity, u_rational, v_rational, u_periodic, v_periodic, u_closed, v_closed, is_trimmed, face_domain_len = payload[
                    idx:idx + 11]
                idx += 11
                face_domain_len = int(face_domain_len)
                face_domain = np.array(payload[idx:idx + face_domain_len])
                idx += face_domain_len

                len_poles = int(payload[idx])
                pshape = [int(val) for val in payload[idx + 1:idx + 4]]
                idx += 4
                poles = np.array(payload[idx:idx + len_poles]).reshape(pshape)
                idx += len_poles

                len_uknots = int(payload[idx])
                idx += 1
                u_knots = np.array(payload[idx:idx + len_uknots])
                idx += len_uknots

                len_vknots = int(payload[idx])
                idx += 1
                v_knots = np.array(payload[idx:idx + len_vknots])
                idx += len_vknots

                len_weights = int(payload[idx])
                wshape = [int(val) for val in payload[idx + 1:idx + 3]]
                idx += 3
                weights = np.array(payload[idx:idx + len_weights]).reshape(wshape)
                # weights_dict = {str(j): weights[j] for j in range(weights.shape[0])}

                res = BSplineSurface(None,
                                      trim_domain=trim_domain,
                                      transform=transform,
                                      u_degree=int(u_degree),
                                      v_degree=int(v_degree),
                                      continuity=int(continuity),
                                      u_rational=bool(u_rational),
                                      v_rational=bool(v_rational),
                                      u_periodic=bool(u_periodic),
                                      v_periodic=bool(v_periodic),
                                      u_closed=bool(u_closed),
                                      v_closed=bool(v_closed),
                                      is_trimmed=bool(is_trimmed),
                                      face_domain=face_domain,
                                      poles=poles,
                                      u_knots=u_knots,
                                      v_knots=v_knots,
                                      weights=weights)

            elif stype == 6:  # Extrusion
                res = Extrusion(None,
                                 trim_domain=trim_domain,
                                 transform=transform,
                                 direction=np.array(payload[idx:idx + 3]),
                                 curve=self._create_3dcurve(payload[idx + 3:]))

            elif stype == 7:  # Revolution
                res = Revolution(None,
                                  trim_domain=trim_domain,
                                  transform=transform,
                                  location=np.array(payload[idx:idx + 3]),
                                  z_axis=np.array(payload[idx + 3:idx + 6]),
                                  curve=self._create_3dcurve(payload[idx + 6:]))

            elif stype == 8:  # Offset
                res = Offset(None,
                              trim_domain=trim_domain,
                              transform=transform,
                              value=np.array(payload[idx]),
                              surface=self._create_surface(payload[idx + 1:]))


            elif stype == 9:  # Other
                res = None
                pass

            return res

        def __init_geometry(self, data, version):
            if version == "3.0":
                curve2d_index = data.get('2dcurves_index', {})[()]
                curve2d_data = data.get('2dcurves', {})[()]
                self.curves2d = [None] * (len(curve2d_index)-1)

                for i in range(len(curve2d_index)-1):
                    tmp = curve2d_data[curve2d_index[i]:curve2d_index[i+1]]
                    ctype = int(tmp[0])
                    interval = tmp[1:3]

                    if ctype == 0:  # Line
                        self.curves2d[i] = Line(None, interval=interval,
                                                location=np.array(tmp[3:5]),
                                                direction=np.array(tmp[5:7]))

                    elif ctype == 1:  # Circle
                        self.curves2d[i] = Circle(None,
                                                interval=interval,
                                                location=np.array(tmp[3:5]),
                                                x_axis=np.array(tmp[5:7]),
                                                y_axis=np.array(tmp[7:9]),
                                                radius=tmp[9])
                    elif ctype == 2:  # Ellipse
                        self.curves2d[i] = Ellipse(None,
                                                interval=interval,
                                                focus1=np.array(tmp[3:5]),
                                                focus2=np.array(tmp[5:7]),
                                                x_axis=np.array(tmp[7:9]),
                                                y_axis=np.array(tmp[9:11]),
                                                maj_radius=tmp[11],
                                                min_radius=tmp[12])
                    elif ctype == 3:  # BSplineCurve
                        degree = int(tmp[3])
                        continuity = int(tmp[4])
                        rational = bool(tmp[5])
                        periodic = bool(tmp[6])
                        closed = bool(tmp[7])
                        len_poles = int(tmp[8])
                        pshape = (int(tmp[9]), int(tmp[10]))
                        poles = np.array(tmp[11:11+len_poles]).reshape(pshape)
                        idx = 11 + len_poles
                        len_knots = int(tmp[idx])
                        knots = tmp[idx+1:idx+1+len_knots]
                        idx = idx + 1 + len_knots
                        len_weights = int(tmp[idx])
                        weights = tmp[idx+1:idx+1+len_weights]

                        self.curves2d[i] = BSplineCurve(None,
                                                        interval=interval,
                                                        degree=degree,
                                                        continuity=continuity,
                                                        rational=rational,
                                                        periodic=periodic,
                                                        closed=closed,
                                                        poles=poles,
                                                        knots=knots,
                                                        weights=weights)
                    elif ctype == 4:  # Other
                        self.curves2d[i] = Other(None, interval=interval)
                    else:
                        raise ValueError(f"Unknown curve type: {ctype}")

                del curve2d_index
                del curve2d_data

                curve3d_index = data.get('3dcurves_index', {})[()]
                curve3d_data = data.get('3dcurves', {})[()]
                self.curves3d = [None] * (len(curve3d_index)-1)
                for i in range(len(curve3d_index)-1):
                    tmp = curve3d_data[curve3d_index[i]:curve3d_index[i+1]]
                    self.curves3d[i] = self._create_3dcurve(tmp)


                del curve3d_index
                del curve3d_data

                surface_index = data.get('surfaces_index', {})[()]
                surface_data = data.get('surfaces', {})[()]
                self.surfaces = [None] * (len(surface_index) - 1)
                surface_type_map = {
                    0: 'Plane',
                    1: 'Cylinder',
                    2: 'Cone',
                    3: 'Sphere',
                    4: 'Torus',
                    5: 'BSpline',
                    6: 'Extrusion',
                    7: 'Revolution',
                    8: 'Offset',
                    9: 'Other'
                }

                for i in range(len(surface_index) - 1):
                    tmp = surface_data[surface_index[i]:surface_index[i + 1]]
                    self.surfaces[i] = self._create_surface(tmp)


                del surface_index
                del surface_data
            else:
                tmp = data.get('2dcurves', {}).values()
                self.curves2d=len(tmp)*[None]
                for curve_data in tmp:
                    index, curve = create_curve(curve_data)
                    self.curves2d[index] = curve

                tmp = data.get('3dcurves', {}).values()
                self.curves3d=len(tmp)*[None]
                for curve_data in tmp:
                    index, curve = create_curve(curve_data)
                    self.curves3d[index] = curve

                tmp = data.get('surfaces', {}).values()
                self.surfaces=len(tmp)*[None]
                for surface_data in tmp:
                    index, surface = create_surface(surface_data)
                    self.surfaces[index] = surface

            self.bbox.append(np.array(data.get('bbox')[:]))

            self.vertices.append(np.array(data.get('vertices')))


    class _topology_data:
        def __init__(self, topology_data, version="3.0"):
            self.edges, self.faces, self.halfedges, self.loops, self.shells, self.solids = [], [], [], [], [], []
            self.__init_topology(topology_data, version)

        def __init_topology(self, data, version):
            if version == "3.0":
                # Edges
                edge_index = data.get('edge_index', {})[()]
                edge_data = data.get('edges', {})[()]
                self.edges = [None] * (len(edge_index)-1)

                for i in range(len(edge_index)-1):
                    tmp = edge_data[edge_index[i]:edge_index[i+1]]
                    local_edge = {'id': i, '3dcurve': tmp[0], 'start_vertex': tmp[1], 'end_vertex': tmp[2]}
                    self.edges[i] = Edge(local_edge)
                # Parallel(n_jobs=-1, backend="threading")(delayed(process_edge)(i) for i in range(len(edge_index)-1))
                del edge_index
                del edge_data

                # Halfedges
                halfedge_index = data.get('halfedge_index', {})[()]
                halfedge_data = data.get('halfedges', {})[()]
                self.halfedges = [None]* (len(halfedge_index)-1)

                for i in range(len(halfedge_index)-1):
                    tmp = halfedge_data[halfedge_index[i]:halfedge_index[i+1]]
                    mates = tmp[3:] if len(tmp) > 3 else []
                    local_halfedge = {'id': i, '2dcurve': tmp[0], 'edge': tmp[1], 'orientation_wrt_edge': bool(tmp[2]), 'mates': mates}
                    self.halfedges[i] = Halfedge(local_halfedge)
                # Parallel(n_jobs=-1, backend="threading")(delayed(process_halfedge)(i) for i in range(len(halfedge_index)-1))
                del halfedge_index
                del halfedge_data

                # Loops
                loop_index = data.get('loop_index', {})[()]
                loop_data = data.get('loops', {})[()]
                self.loops = [None] * (len(loop_index)-1)
                for i in range(len(loop_index)-1):
                    tmp = loop_data[loop_index[i]:loop_index[i+1]]
                    local_loop = {'id': i, 'halfedges': tmp}
                    self.loops[i] = Loop(local_loop)
                # Parallel(n_jobs=-1, backend="threading")(delayed(process_loop)(i) for i in range(len(loop_index)-1))
                del loop_index
                del loop_data

                # Shells
                shell_index = data.get('shell_index', {})[()]
                shell_data = data.get('shells', {})[()]
                self.shells = [None] * (len(shell_index)-1)
                for i in range(len(shell_index)-1):
                    tmp = shell_data[shell_index[i]:shell_index[i+1]]
                    faces = []
                    for j in range(1, len(tmp)-1, 2):
                        faces.append( (tmp[j], bool(tmp[j+1])) )
                    local_shell = {'id': i, 'orientation_wrt_solid': tmp[0], 'faces': faces}
                    self.shells[i] = Shell(local_shell)
                # Parallel(n_jobs=-1, backend="threading")(delayed(process_shell)(i) for i in range(len(shell_index)-1))
                del shell_index
                del shell_data

                # Solids
                solid_index = data.get('solid_index', {})[()]
                solid_data = data.get('solids', {})[()]
                self.solids = [None] * (len(solid_index)-1)
                for i in range(len(solid_index)-1):
                    tmp = solid_data[solid_index[i]:solid_index[i+1]]
                    local_solid = {'id': i, 'shells': tmp}
                    self.solids[i] = TopoSolid(local_solid)
                # Parallel(n_jobs=-1, backend="threading")(delayed(process_solid)(i) for i in range(len(solid_index)-1))
                del solid_index
                del solid_data

                # Faces
                face_index = data.get("face_index")
                if not isinstance(face_index, dict) and not face_index:
                    print('No face is avaialbe in the input - Skipping')
                    return
                face_index = face_index[()]
                face_data = data.get('faces', {})[()]
                self.faces = [None] * (len(face_index)-1)
                SING_WIDTH = 14

                for i in range(len(face_index) - 1):
                    tmp = face_data[face_index[i]:face_index[i + 1]]

                    exact_domain = tmp[:4]
                    has_singularities = bool(tmp[4])
                    nr_singularities = int(tmp[5])

                    # header positions (matches: ed + [hs, ns, os, srf, so] + singularities + loops)
                    outer_loop = int(tmp[6])
                    surface = int(tmp[7])
                    surface_orientation = bool(tmp[8])

                    sing_start = 9
                    sing_end = sing_start + nr_singularities * SING_WIDTH

                    sing_flat = tmp[sing_start:sing_end]

                    singularities = []
                    if has_singularities and nr_singularities > 0:
                        if len(sing_flat) != nr_singularities * SING_WIDTH:
                            raise ValueError(
                                f"Face {i}: expected {nr_singularities * SING_WIDTH} singularity values, got {len(sing_flat)}"
                            )

                        for s in range(nr_singularities):
                            b = s * SING_WIDTH
                            block = sing_flat[b:b + SING_WIDTH]

                            sg = {
                                "first2d": np.array(block[0:2], dtype=np.float64),
                                "firstpar": float(block[2]),
                                "last2d": np.array(block[3:5], dtype=np.float64),
                                "lastpar": float(block[5]),
                                "point2d": np.array(block[6:8], dtype=np.float64),
                                "point3d": np.array(block[8:11], dtype=np.float64),
                                "precision": float(block[11]),
                                "rank": int(block[12]),
                                "uiso": bool(int(block[13])),
                            }
                            singularities.append(sg)

                    loops = np.array(tmp[sing_end:], dtype=np.int64)

                    local_face = {
                        "id": i,
                        "exact_domain": exact_domain,
                        "has_singularities": has_singularities,
                        "nr_singularities": nr_singularities,
                        "outer_loop": outer_loop,
                        "surface": surface,
                        "singularities": singularities,
                        "surface_orientation": surface_orientation,
                        "loops": loops,
                    }
                    self.faces[i] = Face(local_face)
                # Parallel(n_jobs=-1, backend="threading")(delayed(process_face)(i) for i in range(len(face_index)-1))
                del face_index
                del face_data
            else:
                entity_map = {
                    'edges': (self.edges, _get_edges),
                    'faces': (self.faces, _get_faces),
                    'halfedges': (self.halfedges, _get_halfedges),
                    'loops': (self.loops, _get_loops),
                    'shells': (self.shells, Shell),
                    'solids': (self.solids, TopoSolid)
                }

                for entity, (attr_list, constructor) in entity_map.items():
                    entity_data = Topology._get_topo_data(data, entity)
                    attr_list.extend(constructor(item) for item in entity_data)


def _get_edges(edge_data): return Edge(edge_data)
def _get_faces(face_data): return Face(face_data)
def _get_halfedges(halfedge_data): return Halfedge(halfedge_data)
def _get_loops(loop_data): return Loop(loop_data)



"""
Curve classes for parametric curves (Line, Circle, Ellipse, B-spline, etc).
Provide methods to sample points on the curve and compute derivatives.
"""
import numpy as np
from scipy.interpolate import BSpline
from geomdl import NURBS

_CURVE_X, _CURVE_W = np.polynomial.legendre.leggauss(4)

def create_curve(curve_data, compute_index=True):
    """
    Factory function to create a Curve object from HDF5 curve data.
    Returns (index, curve_object). If compute_index is False, index is None.
    """
    if compute_index:
        index = int(curve_data.name.split("/")[-1])
    else:
        index = None
    if isinstance(curve_data, dict):
        curve_type = curve_data.get('type')
    else:
        curve_type = curve_data.get('type')[()].decode('utf-8')
    curve_map = {
        'Line': Line,
        'Circle': Circle,
        'Ellipse': Ellipse,
        'BSpline': BSplineCurve,
        'Other': Other
    }
    curve_class = curve_map.get(curve_type)
    if curve_class:
        return index, curve_class(curve_data)
    else:
        # Unsupported curve type
        return index, None


class Curve:
    """Base class for all curve types."""
    def sample(self, points):
        """Sample points on the curve at given parametric values."""
        raise NotImplementedError("Sample method must be implemented by subclasses")
    def derivative(self, points, order=1):
        """Evaluate the curve's derivative of given order at parametric points."""
        raise NotImplementedError("Derivative method must be implemented by subclasses")
    # def get_length(self):
    #     """Approximate the curve length via Gaussian quadrature (4-point Legendre)."""
    #     if getattr(self, "length", None) is not None and self.length != -1:
    #         return self.length
    #     # Compute length via integration of first derivative
    #     pts, weights = np.polynomial.legendre.leggauss(4)
    #     pts = (pts + 1) * 0.5 * (self.interval[0, 1] - self.interval[0, 0]) + self.interval[0, 0]
    #     deriv = self.derivative(pts[:, None], order=1)
    #     lengths = np.linalg.norm(deriv, axis=1)
    #     self.length = np.sum(lengths * weights) * (self.interval[0, 1] - self.interval[0, 0]) / 2
    #     return self.length
    def get_length(self):

        if getattr(self, "length", None) is not None and self.length != -1:
            return self.length

        t0, t1 = self.interval[0]
        dt = 0.5 * (t1 - t0)
        tm = 0.5 * (t0 + t1)

        t = dt * _CURVE_X + tm

        deriv = self.derivative(t[:, None], order=1)
        speeds = np.linalg.norm(deriv, axis=1)

        self.length = float(np.dot(speeds, _CURVE_W) * dt)
        return self.length

    def __eq__(self, other):

        if type(self) is not type(other):
            return NotImplemented

        rtol = 1e-8
        atol = 1e-10

        def _cmp(v1, v2):
            if isinstance(v1, (np.ndarray, list, tuple)) or isinstance(v2, (np.ndarray, list, tuple)):
                a1 = np.asarray(v1)
                a2 = np.asarray(v2)
                if a1.shape != a2.shape:
                    return False
                if np.issubdtype(a1.dtype, np.number) and np.issubdtype(a2.dtype, np.number):
                    return np.allclose(a1, a2, rtol=rtol, atol=atol)
                return np.array_equal(a1, a2)

            if isinstance(v1, (int, float, bool, np.generic)) and isinstance(v2, (int, float, bool, np.generic)):
                return v1 == v2

            return v1 == v2

        ignore = {"length", "bspline"}
        attrs1 = {k: v for k, v in self.__dict__.items()
                  if not k.startswith("_") and k not in ignore}
        attrs2 = {k: v for k, v in other.__dict__.items()
                  if not k.startswith("_") and k not in ignore}

        if attrs1.keys() != attrs2.keys():
            return False

        for name in attrs1.keys():
            if not _cmp(attrs1[name], attrs2[name]):
                return False

        return True



class Line(Curve):
    """Straight line curve defined by a start location and direction vector."""
    def __init__(self, line, location=None, interval=None, direction=None, transform=None):
        self.length = -1

        if line is None:
            self.location = np.array(location).reshape(-1, 1).T
            self.interval = np.array(interval).reshape(-1, 1).T
            self.direction = np.array(direction).reshape(-1, 1).T
            if transform is not None:
                self.transform = transform
            self.shape_name = 'Line'
        else:
            self.location = np.array(line.get('location')[()]).reshape(-1, 1).T
            self.interval = np.array(line.get('interval')[()]).reshape(-1, 1).T
            self.direction = np.array(line.get('direction')[()]).reshape(-1, 1).T
            if self.direction.shape[1] == 3:
                self.transform = np.array(line.get('transform')[()])  # transformation matrix if present
            self.shape_name = line.get('type')[()].decode('utf8')

    def sample(self, sample_points):
        if sample_points.size == 0:
            return self.location
        # Parametric coordinate in interval [0,1] maps to location + t * direction
        return self.location + sample_points * self.direction

    def get_length(self):
        if self.length == -1:
            self.length = np.linalg.norm((self.interval[0, 1] - self.interval[0, 0]) * self.direction)
        return self.length

    def derivative(self, sample_points, order=1):
        if order == 1:
            # First derivative of line is constant direction
            return np.tile(self.direction, (sample_points.shape[0], 1))
        # Higher derivatives of a line are zero
        return np.zeros((sample_points.shape[0], self.location.shape[1]))


class Circle(Curve):
    """Circular arc or full circle curve."""
    def __init__(self, circle, location=None, radius=None, interval=None, x_axis=None, y_axis=None, z_axis=None, transform=None):
        self.length = -1

        if circle is None:
            self.location = np.array(location).reshape(-1, 1).T
            self.radius = radius
            self.interval = np.array(interval).reshape(-1, 1).T
            self.x_axis = np.array(x_axis).reshape(-1, 1).T
            self.y_axis = np.array(y_axis).reshape(-1, 1).T
            if z_axis is not None:
                self.z_axis = np.array(z_axis).reshape(-1, 1).T
            if transform is not None:
                self.transform = transform
            self.shape_name = 'Circle'
        else:
            self.location = np.array(circle.get('location')[()]).reshape(-1, 1).T
            self.radius = float(circle.get('radius')[()])
            self.interval = np.array(circle.get('interval')[()]).reshape(-1, 1).T
            self.x_axis = np.array(circle.get('x_axis')[()]).reshape(-1, 1).T
            self.y_axis = np.array(circle.get('y_axis')[()]).reshape(-1, 1).T
            if 'z_axis' in circle:
                self.z_axis = np.array(circle.get('z_axis')[()]).reshape(-1, 1).T
                self.transform = np.array(circle.get('transform')[()])
            self.shape_name = circle.get('type')[()].decode('utf8')

    def sample(self, sample_points):
        if sample_points.size == 0:
            return self.location
        # parametric angle values in sample_points
        circle_points = self.location + self.radius * (np.cos(sample_points) * self.x_axis + np.sin(sample_points) * self.y_axis)
        return circle_points

    def derivative(self, sample_points, order=1):
        if order == 0:
            return self.sample(sample_points)
        # Circle derivatives are periodic every 4th derivative
        if order % 4 == 0 and order != 0:
            # Cosine form (same as first derivative for order 4)
            return self.radius * (np.cos(sample_points) * self.x_axis + np.sin(sample_points) * self.y_axis)
        elif order % 4 == 1:
            return -self.radius * np.sin(sample_points) * self.x_axis + self.radius * np.cos(sample_points) * self.y_axis
        elif order % 4 == 2:
            return -self.radius * (np.cos(sample_points) * self.x_axis + np.sin(sample_points) * self.y_axis)
        else:  # order % 4 == 3
            return self.radius * (np.sin(sample_points) * self.x_axis - np.cos(sample_points) * self.y_axis)


class Ellipse(Curve):
    """Elliptical curve (full ellipse or arc)."""
    def __init__(self, ellipse, focus1=None, focus2=None, interval=None, maj_radius=None, min_radius=None, x_axis=None, y_axis=None, z_axis=None, transform=None):
        self.length = -1

        if ellipse is None:
            self.focus1 = np.array(focus1).reshape(-1, 1).T
            self.focus2 = np.array(focus2).reshape(-1, 1).T
            self.interval = np.array(interval).reshape(-1, 1).T
            self.maj_radius = maj_radius
            self.min_radius = min_radius
            self.x_axis = np.array(x_axis).reshape(-1, 1).T
            self.y_axis = np.array(y_axis).reshape(-1, 1).T
            if z_axis is not None:
                self.z_axis = np.array(z_axis).reshape(-1, 1).T
            if transform is not None:
                self.transform = transform
            self.shape_name = 'Ellipse'
        else:
            self.focus1 = np.array(ellipse.get('focus1')[()]).reshape(-1, 1).T
            self.focus2 = np.array(ellipse.get('focus2')[()]).reshape(-1, 1).T
            self.interval = np.array(ellipse.get('interval')[()]).reshape(-1, 1).T
            self.maj_radius = float(ellipse.get('maj_radius')[()])
            self.min_radius = float(ellipse.get('min_radius')[()])
            self.x_axis = np.array(ellipse.get('x_axis')[()]).reshape(-1, 1).T
            self.y_axis = np.array(ellipse.get('y_axis')[()]).reshape(-1, 1).T

            self.shape_name = ellipse.get('type')[()].decode('utf8')
            if 'z_axis' in ellipse:
                self.z_axis = np.array(ellipse.get('z_axis')[()]).reshape(-1, 1).T
                self.transform = np.array(ellipse.get('transform')[()])  # 4x4 transform matrix if present

        # Center of ellipse
        self.center = (self.focus1 + self.focus2) / 2

    def sample(self, sample_points):
        # Parametric angle values in sample_points
        ellipse_points = self.center + self.maj_radius * np.cos(sample_points) * self.x_axis + self.min_radius * np.sin(sample_points) * self.y_axis
        return ellipse_points

    def derivative(self, sample_points, order=1):
        if order % 4 == 0:
            if order == 0:
                return self.sample(sample_points)
            return self.maj_radius * np.cos(sample_points) * self.x_axis + self.min_radius * np.sin(sample_points) * self.y_axis
        elif order % 4 == 1:
            return -self.maj_radius * np.sin(sample_points) * self.x_axis + self.min_radius * np.cos(sample_points) * self.y_axis
        elif order % 4 == 2:
            return -self.maj_radius * np.cos(sample_points) * self.x_axis - self.min_radius * np.sin(sample_points) * self.y_axis
        else:
            return self.maj_radius * np.sin(sample_points) * self.x_axis - self.min_radius * np.cos(sample_points) * self.y_axis

class BSplineCurve(Curve):
    def __init__(self, bspline, closed=None, degree=None, continuity=None, poles=None, knots=None, weights=None, interval=None, rational=None, periodic=None, transform=None):
        self.length = -1

        if bspline is None:
            self.closed = bool(closed)
            self.degree = int(degree)
            self.continuity = int(continuity)
            self.poles = np.array(poles)
            self.knots = np.array(knots).reshape(-1, 1).T
            self.weights = np.array(weights).reshape(-1, 1)
            self.interval = np.array(interval).reshape(-1, 1).T
            self.rational = bool(rational)
            self.periodic = bool(periodic)
            if transform is not None:
                self.transform = transform
            self.shape_name = 'BSpline'
        else:
            self.closed = bool(bspline.get('closed')[()])
            self.degree = int(bspline.get('degree')[()])
            self.continuity = int(bspline.get('continuity')[()])
            self.poles = np.array(bspline.get('poles')[()])
            self.knots = np.array(bspline.get('knots')[()]).reshape(-1, 1).T
            self.weights = np.array(bspline.get('weights')[()]).reshape(-1, 1)
            self.interval = np.array(bspline.get('interval')[()]).reshape(-1, 1).T
            self.rational = bool(bspline.get('rational')[()])
            self.periodic = bool(bspline.get('periodic')[()])
            self.shape_name = bspline.get('type')[()].decode('utf8')
            if self.poles.shape[1] == 3:
                self.transform = np.array(bspline.get('transform')[()])

        # Create BSpline or NURBS curve object
        if self.rational:
            self.bspline = NURBS.Curve(normalize_kv=False)
            self.bspline.degree = self.degree
            self.bspline.ctrlpts = self.poles.tolist()
            self.bspline.knotvector = self.knots.flatten().tolist()
            self.bspline.weights = self.weights.flatten().tolist()
        else:
            self.bspline = BSpline(self.knots.T[:,0], self.poles, self.degree)

    def sample(self, sample_points):
        if self.rational:
            return np.array(self.bspline.evaluate_list(sample_points.flatten().tolist()))

        return np.squeeze(self.bspline(sample_points))

    def derivative(self, sample_points, order=1):

        if order == 0:
            return self.sample(sample_points)

        elif self.degree < order:
            return np.zeros([sample_points.shape[0], self.poles.shape[1]])
        else:
            if self.rational:
                res = np.zeros([sample_points.shape[0], self.poles.shape[1]])
                for i in range(sample_points.shape[0]):
                    d = self.bspline.derivatives(sample_points[i, 0], order)
                    res[i, :] = d[-1]
                return res
            b_spline_derivative = self.bspline.derivative(order)
            return np.squeeze(b_spline_derivative(sample_points))

class Other(Curve):
    """Fallback for unsupported curve types."""
    def __init__(self, other, interval=None, transform=None):
        if other is None:
            self.shape_name = 'Other'
            self.interval = np.array(interval).reshape(-1, 1).T
            if transform is not None:
                self.transform = transform
        else:
            self.shape_name = other.get('type')[()].decode('utf8')
            self.interval = np.array(other.get('interval')[()]).reshape(-1, 1).T

    def sample(self, sample_points):
        # Unsupported: return empty array
        return np.array([]).reshape(0, 3)

    def derivative(self, sample_points, order=1):
        return np.array([]).reshape(0, 3)

    def get_length(self):
        return 0.0





"""
Functions for processing Shape parts: sampling points and computing normals.
"""
import numpy as np
from abs import sampler
from abspy import poisson_disk_downsample
# from abspy import poisson_grid_downsample

def _slice(res, indices, n):
    if isinstance(res, list) or isinstance(res, tuple):
        assert(len(res) == n)
        return [res[i] for i in range(n) if indices[i]]

    assert (isinstance(res, np.ndarray))
    return res[indices]

def estimate_total_surface_area(part):
    """Estimate the total surface area of all faces in a Shape part."""
    total_area = 0.0
    for face in part.faces:
        total_area += face.get_area()
    return total_area

def estimate_total_curve_length(part):
    """Estimate the total curve length of all edges in a Shape part."""
    total_length = 0.0
    for edge in part.edges:
        total_length += edge.get_length()
    return total_length


def process_part(part,
                 num_samples,
                 face_func,
                 edge_func,
                 points_ratio,
                 apply_transform,
                 uniform_sample,
                 use_poisson,
                 force_num_points,
                 sample_num_tolerance):
    """
    FIXME
    Sample points on all surfaces and curves of a shape part according to specified number.
    Uses an iterative strategy with oversampling (points_ratio) and Poisson disk downsampling to refine points.
    The lambda_func is a function(part, topo_entity, param_points) that returns associated values (e.g., normals).
    """
    initial_num_points = points_ratio * num_samples
    num_points = initial_num_points
    total_area = estimate_total_surface_area(part)
    total_length = estimate_total_curve_length(part)
    n_face_points = 0
    while True:
        current_pts = []
        current_ss = []

        if face_func is not None:
            for face in part.faces:
                n_surf = int(np.ceil((face.get_area() / total_area) * num_points))
                if uniform_sample:
                    uv_points, pt = sampler.uniform_sample(face, n_surf, min_pts=2)
                else:
                    uv_points, pt = sampler.random_sample(face, n_surf, min_pts=2)
                s = face_func(face, uv_points)
                if s is not None:
                    if apply_transform:
                        # Transform points to surface's local coordinate system
                        R = np.linalg.inv(face.surface.transform[:, :3])
                        t = face.surface.transform[:, 3]
                        pt = (pt - t) @ R.T
                    # Filter out points outside trimming loops
                    index = face.filter_outside_points(uv_points)
                    current_pts.append(pt[index, :])
                    if isinstance(s, list) or isinstance(s, tuple):
                        current_ss += _slice(s, index, pt.shape[0])
                    else:
                        current_ss.append(_slice(s, index, pt.shape[0]))
        if len(current_pts) >0:
            n_face_points = np.concatenate(current_pts, axis=0).shape[0]

        if edge_func is not None:
            for edge in part.edges:
                if edge.curve3d is None or edge.curve3d.shape_name == "Other":
                    continue
                n_edge = int(np.ceil((edge.get_length() / total_length) * num_points))
                if uniform_sample:
                    uv_points, pt = sampler.uniform_sample(edge, n_edge, min_pts=2)
                else:
                    uv_points, pt = sampler.random_sample(edge, n_edge, min_pts=2)
                s = edge_func(edge, uv_points)
                if s is not None:
                    if apply_transform:
                        R = np.linalg.inv(edge.curve3d.transform[:, :3])
                        t = edge.curve3d.transform[:, 3]
                        pt = (pt - t) @ R.T
                    # Edges have no trimming curves; include all sampled points
                    index = np.ones(uv_points.shape[0], dtype=bool)
                    current_pts.append(pt[index, :])
                    if isinstance(s, list) or isinstance(s, tuple):
                        current_ss += _slice(s, index, pt.shape[0])
                    else:
                        current_ss.append(_slice(s, index, pt.shape[0]))
        # Combine all collected points and associated data
        if len(current_pts) == 0:
            pts = np.zeros((0, 3))
            ss = []
        else:
            pts = np.concatenate(current_pts, axis=0)
            try:
                ss = np.concatenate(current_ss, axis=0)
            except ValueError:
                ss = current_ss
        # Stop when enough points collected or nothing collected
        if not force_num_points or pts.shape[0] >= initial_num_points or pts.shape[0] == 0:
            break
        else:
            # Increase num_points for next iteration (upscale by factor to approach initial count)
            num_points = np.ceil(num_points * initial_num_points / max(len(pts), 1) * 1.2)

    # If no points at all, return empty structures
    if pts.shape[0] == 0:
        return pts, pts, pts, pts
    # Poisson disk downsample to exactly num_samples points

    if use_poisson:
        indices = poisson_disk_downsample(pts, num_samples, sample_num_tolerance=sample_num_tolerance)
        # indices = poisson_grid_downsample(pts, num_samples)
    else:
        indices = np.arange(pts.shape[0])

    if force_num_points:
        if len(indices) < num_samples:
            remaining_idx = [i for i in range(len(pts)) if i not in indices]
            additional_indices = np.random.choice(remaining_idx, num_samples - len(indices), replace=False)
            indices = np.concatenate([indices, additional_indices])
        elif len(indices) > num_samples:
            indices = np.random.choice(indices, num_samples, replace=False)

    indices = np.sort(indices)

    face_indices = indices[indices < n_face_points]
    edge_indices = indices[indices >= n_face_points]

    if isinstance(ss, list):
        return pts[face_indices], [ss[i] for i in face_indices], pts[edge_indices], [ss[i] for i in edge_indices]
    else:
        assert(isinstance(ss, np.ndarray))
        return pts[face_indices], ss[face_indices], pts[edge_indices], ss[edge_indices]


def sample_parts(parts,
                 num_samples,
                 face_func=None,
                 edge_func=None,
                 points_ratio=5,
                 apply_transform=True,
                 uniform_sample=False,
                 use_poisson=True,
                 force_num_points=True,
                 sample_num_tolerance=0.04):

    """
        Process a list of parts by sampling each part and returning lists of points and values.
    """
    pts_listf = []
    pts_liste = []
    ss_listf = []
    ss_liste = []
    for part in parts:
        ptsf, ssf, ptse, sse = process_part(part, num_samples, face_func, edge_func, points_ratio=points_ratio,apply_transform=apply_transform, uniform_sample=uniform_sample, use_poisson=use_poisson, force_num_points=force_num_points,sample_num_tolerance=sample_num_tolerance)
        pts_listf.append(ptsf)
        pts_liste.append(ptse)
        ss_listf.append(ssf)
        ss_liste.append(sse)
    return pts_listf, ss_listf, pts_liste, ss_liste




from pathlib import Path
import meshio as mio
import numpy as np
import h5py
from .shape import Shape


def read_parts(file_path):
    """Read all parts from an HDF5 file and construct Shape objects."""
    f = h5py.File(file_path, 'r')
    part = f['parts'].values()
    version = f['parts'].attrs.get('version')
    parts = []
    for i, p in enumerate(part):
        s = Shape(p['geometry'], p['topology'], version)
        parts.append(s)
    return parts


def read_meshes(file_path):
    """Read pre-computed mesh data (points and triangles) for each face of each part from an HDF5 file."""
    f = h5py.File(file_path, 'r')
    part = f['parts'].values()
    version = f['parts'].attrs.get('version')
    meshes = []
    for i, p in enumerate(part):
        if version == '2.0':
            s = Shape(p['geometry'], p['topology'], version)
            if not hasattr(s, 'faces') or not s.faces:
                meshes.append([])
                continue

            mesh_group = p['mesh']
            current_mesh = [None] * len(s.faces)
            for key in mesh_group:
                submesh = mesh_group[key]
                vertices = submesh['points']
                faces = submesh['triangle']
                current_mesh[int(key)] = {
                    'points': vertices,
                    'triangle': faces
                }
            meshes.append(current_mesh)
        elif version == '3.0':
            mesh_group = p['mesh']
            all_points = mesh_group['points'][()]
            all_tris = mesh_group['triangles'][()]
            p_idx = mesh_group['point_index'][()]
            t_idx = mesh_group['triangle_index'][()]
            n_meshes = len(p_idx) - 1
            assert(len(p_idx) == len(t_idx))
            current_mesh = [None] * n_meshes
            for mi in range(n_meshes):
                ps = all_points[p_idx[mi]:p_idx[mi + 1]].reshape(-1, 3).astype(np.float32, copy=False)
                ts = all_tris[t_idx[mi]:t_idx[mi + 1]].reshape(-1, 3).astype(np.int32, copy=False)
                current_mesh[mi] = {
                    'points': ps,
                    'triangle': ts
                }
            meshes.append(current_mesh)
    return meshes

# -------------------------
# Saving utilities
# -------------------------

def save_obj_points(filename, pts):
    """Save a set of 2D/3D points to an .obj file."""
    with open(filename, "w") as f:
        if pts.shape[1] == 2:
            for i in range(pts.shape[0]):
                f.write(f"v {pts[i, 0]} {pts[i, 1]} 0\n")
        else:
            for i in range(pts.shape[0]):
                f.write(f"v {pts[i, 0]} {pts[i, 1]} {pts[i, 2]}\n")

def save_ply_mesh(filename, pts, faces):
    mio.write_points_cells(filename, pts, [("triangle", faces)])

def save_obj_mesh(filename, pts, faces):
    """Save a set of 3D points and faces to an .obj file."""
    if pts.shape[0] == 0:
        print("Skipping saving meshes: mesh is empty")
        return
    with open(filename, "w") as f:
        if pts.shape[1] == 2:
            for i in range(pts.shape[0]):
                f.write(f"v {pts[i, 0]} {pts[i, 1]} 0\n")
        else:
            for i in range(pts.shape[0]):
                f.write(f"v {pts[i, 0]} {pts[i, 1]} {pts[i, 2]}\n")
        for i in range(faces.shape[0]):
            f.write(f"f {faces[i, 0] + 1} {faces[i, 1] + 1} {faces[i, 2] + 1}\n")



def save_ply(filename, P, normals=None):
    '''
    Save a set of 3D points to a .ply file. Optionally, also save normals.
    '''
    new_pts = np.vstack(P)

    if normals:
        new_normal = np.vstack(normals)

        if new_pts.shape[0] != new_normal.shape[0]:
            raise ValueError("The number of points and normals must be the same")

        if new_pts.shape[1] != 3 or new_normal.shape[1] != 3:
            raise ValueError("Both pts and normals must have shape (n, 3)")
    else:
        if new_pts.shape[1] != 3:
            raise ValueError("Points must have shape (n, 3)")

    # mio.write_points_cells(filename, new_pts, [("vertex", np.arange(new_pts.shape[0], dtype=np.int32)[:, None])])

    new_pts = np.asarray(new_pts, dtype=np.float64)
    mesh = mio.Mesh(points=new_pts, cells=[])
    mio.write(filename, mesh, file_format="ply", binary=True)

def save_to_xyz(points, filename):
    with open(filename, 'w') as f:
        for point in points:
            f.write(f"{point[0]} {point[1]} {point[2]}\n")


def save_vtu(save_file_path , P):
    m = mio.Mesh(P, cells={"triangle":np.array([np.arange(P.shape[0]), np.arange(P.shape[0]), np.arange(P.shape[0])]).T})
    m.write(save_file_path)

def get_mesh(meshes):
    global_vertices = []
    global_faces = []
    vertex_offset = 0
    for mesh in meshes:
        for sub_mesh in mesh:
            if sub_mesh is None:
                continue
            vertices = sub_mesh["points"][:]
            if len(vertices) == 0:
                continue
            global_vertices.append(vertices)
            faces = sub_mesh["triangle"][:] + vertex_offset
            global_faces.append(faces)
            vertex_offset += vertices.shape[0]
    if global_vertices:
        global_vertices = np.vstack(global_vertices)
    else:
        global_vertices = np.empty((0, 3))
    if global_faces:
        global_faces = np.vstack(global_faces)
    else:
        global_faces = np.empty((0, 3), dtype=int)
    return global_vertices, global_faces

def get_mesh_per_part(meshes):

    global_vertices = []
    global_faces = []
    for mesh in meshes:
        V_list = []
        F_list = []
        offset = 0
        for sub_mesh in mesh:
            if sub_mesh is None:
                continue
            vertices = np.asarray(sub_mesh["points"], dtype=np.float32).reshape(-1, 3)
            faces = np.asarray(sub_mesh["triangle"], dtype=np.int32).reshape(-1, 3)
            if vertices.shape[0] == 0 or faces.shape[0] == 0:
                continue
            V_list.append(vertices)
            F_list.append(faces + offset)
            offset += vertices.shape[0]
        if V_list:
            Vp = np.vstack(V_list)
            Fp = np.vstack(F_list) if F_list else np.zeros((0, 3), dtype=np.int32)
        else:
            Vp = np.zeros((0, 3), dtype=np.float32)
            Fp = np.zeros((0, 3), dtype=np.int32)
        global_vertices.append(Vp)
        global_faces.append(Fp)
    return global_vertices, global_faces



from abs.utils import *
from abs.part_processor import *
import unittest
import os


def face_func(face, points):
    return face.normal(points)


def curve_func(edge, points):
    return None

class TestShapeFunctions(unittest.TestCase):

    def test_get_parts_integration(self):
        name = 'Cone'
        sample_name = f'{name}.hdf5'
        file_path = os.path.join(os.path.dirname(__file__), '..', 'data', 'sample_hdf5', sample_name)
        file_path = os.path.normpath(file_path)

        num_samples = 5000
        parts = read_parts(file_path)
        face_points, face_feat, edge_points, edge_feat = sample_parts(parts, num_samples, face_func, curve_func)

        self.assertEqual(len(face_points), len(parts))
        self.assertEqual(len(face_feat), len(parts))
        self.assertEqual(len(face_points), len(face_feat))

if __name__ == '__main__':
    unittest.main()



from abs.part_processor import *
from abs.utils import *
import unittest
from test_utilities import *


def face_func(face, points):
    return face.normal(points)


def curve_func(edge, points):
    return None


class TestApplication(unittest.TestCase):
    def test_run_application(self):
        file_path = get_file("Ellipse.hdf5")
        num_samples = 4000
        parts = read_parts(file_path)
        meshes = read_meshes(file_path)

        face_points, face_feat, edge_points, edge_feat = sample_parts(parts, num_samples, face_func,
                                                                      curve_func)
        V, F = get_mesh(meshes)

